# Copyright 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import binascii
import ctypes
from ctypes import (
    CFUNCTYPE, Structure, c_uint8, c_uint32, c_long, c_void_p,
    c_uint64, c_char_p, c_int, c_uint16, POINTER, Union, byref)
import json
import random
import signal
import subprocess
import sys
import time
import yaml

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import (
    WjhDropRawInfo,
    WjhL1DropAggregate,
    WjhDropAggregate)
from scapy.all import (raw, Ether, IP, IPv6, TCP, UDP, VXLAN)
from netq_lib.common.utils import subprocess_wrapper
from netq_lib.common.enums import WjhConfKw

sys.path.append('/usr/lib/cumulus/wjh/')
from wjh_types import *

AGENTSTATUS_SUCCESS = 0
AGENTSTATUS_FAILURE = 1

WJH_DROP_REASON_ID_NON_ROUTABLE_PACKET_E = 301
WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX0_16_E = 305
WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX1_16_E = 306
WJH_DROP_REASON_ID_NON_IP_PACKET_E = 307

WJH_DROP_REASON_ID_MLAG_PORT_ISOLATION_E = 201
WJH_DROP_REASON_ID_INGRESS_SPANNING_TREE_FILTER_E = 205

default_filters = [WJH_DROP_REASON_ID_NON_ROUTABLE_PACKET_E,
                   WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX0_16_E,
                   WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX1_16_E,
                   WJH_DROP_REASON_ID_NON_IP_PACKET_E,
                   WJH_DROP_REASON_ID_MLAG_PORT_ISOLATION_E,
                   WJH_DROP_REASON_ID_INGRESS_SPANNING_TREE_FILTER_E]

WJH_MIRROR_CONG_INVALID = 0xFFFF
WJH_MIRROR_TCLASS_INVALID = 0x1F
WJH_MIRROR_LATENCY_INVALID = 0xFFFFFF
WJH_MIRROR_LATENCY_NSEC_UNIT = 64
WJH_BUFFER_OCCUPANCY_UNIT = 8

client_instance = None

WJH_LIB_XML_PATH = '/etc/netq/wjh_lib_conf.xml'
WJH_TELE_PATH = '/usr/sbin/netq-wjh-tele'
WJH_DUMMY_DATA_PATH = '/etc/netq/.wjh.data'
MLX_PLATFORM_JSON_PATH = '/usr/share/cumulus-platform/common/platform-info/mellanox.json'
SWITCHD_SW_TABLE_PATH = '/cumulus/switchd/run/software-tables'
SDK_PID_LAST = '/var/run/cumulus/sx_sdk.last'
THRESHOLDS_LAST = '/var/run/cumulus'
CL_PORTTAB = '/var/lib/cumulus/porttab'
CL_PORTTAB_LAST = '/var/run/cumulus/porttab.last'

severity_to_str = {
    WJH_SEVERITY_NOTICE_E: 'Notice',
    WJH_SEVERITY_WARNING_E: 'Warning',
    WJH_SEVERITY_ERROR_E: 'Error'
}

acl_bp_to_str = {
    WJH_ACL_BINDING_POINT_INGRESS: 'Ingress ACL',
    WJH_ACL_BINDING_POINT_EGRESS: 'Egress ACL',
    WJH_ACL_BINDING_POINT_RIF_INGRESS: 'RIF Ingress ACL',
    WJH_ACL_BINDING_POINT_RIF_EGRESS: 'RIF Egress ACL'
}


class WjhClient(object):

    def __init__(self, conf, logger, platform, is_mlx=True):
        self._conf = conf
        self._logger = logger
        self.wjh_lib = None
        self.portmap = {}
        self.lagmap = {}
        self.lidmap = {}
        self.agg_cache = {}
        self.sdk_pid = 0
        self.port_map_id = 0
        self.lag_map_id = 0
        self.is_mlx = is_mlx
        self.mlx_platform = platform
        self.need_tele_init = False

        signal.signal(signal.SIGTERM, sig_handler)

    def create_user_channel(self, channel_type):
        user_channel = c_uint32()
        if self.wjh_lib.wjh_user_channel_create(channel_type,
                                                byref(user_channel)) != 0:
            self._logger.error("Failed to create user channel")
        else:
            self._logger.info("create channel id %d" % (user_channel.value))

        return user_channel.value


    def get_counter(self, user_channel_id):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_counter_get(user_channel_id,
                    byref(drop_counter), 0) != 0:

            self._logger.error("Failed to get counter")
        else:
            self._logger.info("channel[%d]: drop_counter [%d], recieve_counter [%d]" % (
                              user_channel_id, drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def get_global_counter(self):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_global_counter_get(byref(drop_counter), 0) != 0:
            self._logger.error("Failed to get global counter")
        else:
            self._logger.error("global: drop_counter [%d], recieve_counter [%d]" % (
                               drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def configure_drop_reason_group(self, drop_reason_group, channel_id):
        if self.wjh_lib.wjh_drop_reason_group_bind(drop_reason_group, channel_id) != 0:
            self._logger.error("Failed to bind drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_enable(drop_reason_group, WJH_SEVERITY_ALL_E) != 0:
            self._logger.error("Failed to enable drop reason group {}".format(drop_reason_group))


    def deconfigure_drop_reason_group(self, drop_reason_group):
        if self.wjh_lib.wjh_drop_reason_group_disable(drop_reason_group, WJH_SEVERITY_ALL_E) != 0:
            self._logger.error("Failed to disable drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_unbind(drop_reason_group) != 0:
            self._logger.error("Failed to unbind drop reason group {}".format(drop_reason_group))

        if self.wjh_lib.wjh_drop_reason_group_deinit(drop_reason_group) != 0:
            self._logger.error("Failed to deinit drop reason group {}".format(drop_reason_group))


    def process_threshold_cfg(self, thresh, th_type):
        tele = WJH_TELE_PATH
        th_matrix = []
        old_th_matrix = []

        # get the previous settings
        old_th_file = '{}/{}.last'.format(THRESHOLDS_LAST, th_type)

        if not self.need_tele_init:
            try:
                with open(old_th_file, 'r') as json_file:
                    old_th_matrix = json.load(json_file)
            except Exception as e:
                pass

        for i in range(0, 8):
            th_matrix.append({'enabled': False, 'tc': i, 'ports': [], 'hi': 0, 'lo': 0})

        if thresh:
            #self._logger.info(thresh)
            for ths in thresh:
                tcs = ths.get('tc')
                if tcs is None:
                    continue

                ports = []
                if ths.get('ports'):
                    for p in ths.get('ports').split(','):
                        lid = self.lidmap.get(p)
                        if not lid:
                            continue
                        ports.append(lid)
                else:
                    ports = self.portmap.keys()

                tcs = str(tcs) if isinstance(tcs, int) else tcs
                for a_tc in tcs.split(','):
                    r = a_tc.split('-')
                    s = int(r[0])
                    e = s
                    if len(r) > 1:
                        e = int(r[1])
                    if s > 7 or e > 7:
                        continue

                    for i in range(s, e + 1):
                        th_matrix[i]['enabled'] = True
                        th_matrix[i]['hi'] = ths.get('hi', 0)
                        th_matrix[i]['lo'] = ths.get('lo', 0)
                        th_matrix[i]['ports'] = ports

                        # the left over ports for the tc in the old setting will be cleaned up
                        if len(old_th_matrix) > i and old_th_matrix[i]['enabled']:
                            for p in ports:
                                if p in th_matrix[i]['ports'] and p in old_th_matrix[i]['ports']:
                                    old_th_matrix[i]['ports'].remove(p)

        for ths in th_matrix:
            for lid in self.portmap.keys():
                try:
                    sub_cmd = ''
                    if lid not in ths['ports'] or not ths['enabled']:
                        continue

                    cmd = '{} {} --port {} --tc {} --type {}'.format(tele, '--get',
                          lid, ths['tc'], th_type)
                    output = subprocess_wrapper(cmd.split())
                    if output:
                        for l in output.splitlines():
                            if 'sx_api_tele_threshold_get' in l:
                                _, _, rc = l.split()
                                if rc == '0':
                                    sub_cmd = '--update'
                                else:
                                    sub_cmd = ''
                                break

                    cmd = '{} {} --port {} --tc {} --type {} --threshold_hi {} --threshold_lo {}'.format(
                            tele, sub_cmd, lid, ths['tc'], th_type, ths['hi'], ths['lo'])
                    self._logger.debug(cmd)
                    subprocess_wrapper(cmd.split())
                except subprocess.CalledProcessError as e:
                    self._logger.error(e)

        # clean up the old settings that are removed
        for ths in old_th_matrix:
            if ths['enabled'] and len(ths['ports']) > 0:
                for p in ths['ports']:
                    try:
                        cmd = '{} {} --port {} --tc {} --type {} --threshold_hi 0 --threshold_lo 0'.format(
                              tele, '--clear', p, ths['tc'], th_type)

                        output = subprocess_wrapper(cmd.split())
                        self._logger.debug(cmd)
                    except subprocess.CalledProcessError as e:
                        self._logger.error(e)

        # save the threshold settings
        try:
            with open(old_th_file, 'w') as json_file:
                json.dump(th_matrix, json_file)
        except Exception as e:
            pass


    def init_latency_and_congestion_thresholds(self):
        # NOTE: this should really be done in CL as switchd init
        #       for now, the config is put under agent as follows
        # netq-agent:
        #     log_level: info
        #     wjh: true
        #     wjh_latency_thresholds:
        #         - tc: 0-4,7
        #           ports: swp5,swp6
        #           hi: 1
        #           lo: 0
        #         - tc: 5
        #           hi: 2
        #           lo: 1
        #     wjh_congestion_thresholds:
        #         - tc: 4
        #           ports: swp5,swp6
        #           hi: 1
        #           lo: 0

        tele = WJH_TELE_PATH
        if self.need_tele_init:
            cmd = '{} --init'.format(tele)
            try:
                subprocess_wrapper(cmd.split())
            except subprocess.CalledProcessError as e:
                self._logger.error(e)
            self._logger.info('Initialize sx tele module')

        lat_thresh = self._conf.get('netq-agent').get('wjh_latency_thresholds', False)
        if lat_thresh:
            self.process_threshold_cfg(lat_thresh, 'latency')


        cng_thresh = self._conf.get('netq-agent').get('wjh_congestion_thresholds', False)
        if cng_thresh:
            self.process_threshold_cfg(cng_thresh, 'congestion')


    def latency_monitor_supported(self):
        cmd = '/usr/bin/dpkg -l sx-sdk-eth'
        output = subprocess_wrapper(cmd.split())
        for l in output.splitlines():
            if 'sx-sdk-eth' in l:
                pkg = l.split()[2]
                sdk_ver = pkg.split('.')
                mj = int(sdk_ver[2])
                md = int(sdk_ver[3])
                mi = int(sdk_ver[4].split('-')[0])
                self._logger.info('mlx sdk version: {}.{}.{}'.format(mj, md, mi))
                if mj == 4:
                    if (md == 3 and mi >= 3206) or (md >= 4):
                        self._logger.info('Platform supports latency and congestion threshold monitoring')
                        return True
                elif mj > 4:
                    return True
        return False

    def buffer_cb_supported(self):
        plat_name = self.mlx_platform.split()[1]
        self._logger.info('PLATFORM: {}'.format(plat_name))
        path = MLX_PLATFORM_JSON_PATH
        try:
            with open(path, 'r') as f:
                plat = json.load(f)
                if plat.get(plat_name):
                    plat_json = plat[plat_name][0]
                    asic = plat_json['soc']['model']
                    if asic != 'Spectrum':
                        self._logger.info('Platform supports wjh buffer callback')
                        return True
        except OSError:
            pass

        return False


    def wjh_callback(self, channel_id, aggregation_channel_id):
        global l2_cb, router_cb, tunnel_cb, acl_cb, l1_cb, buffer_cb

        if self.get_switchd_mapping_tables() == AGENTSTATUS_FAILURE:
            return

        if self.get_port_map() == AGENTSTATUS_FAILURE:
            return

        # L2
        l2_cb = WjhDropCallbacks()
        l2_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L2_E
        l2_cb.raw_cb.L2 = WjhL2DropRawCb(self.wjh_l2_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L2_E, byref(attr), byref(l2_cb)) != 0:
            self._logger.error("Failed to init L2 drop reason group")

        # L3
        router_cb = WjhDropCallbacks()
        router_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ROUTER_E
        router_cb.raw_cb.router = WjhRouterDropRawCb(self.wjh_router_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ROUTER_E, byref(attr), byref(router_cb)) != 0:
            self._logger.error("Failed to init router drop reason group")

        # tunnel
        tunnel_cb = WjhDropCallbacks()
        tunnel_cb.drop_reason_group = WJH_DROP_REASON_GROUP_TUNNEL_E
        tunnel_cb.raw_cb.tunnel = WjhTunnelDropRawCb(self.wjh_tunnel_drop_raw_cb)
        attr = WjhDropReasonGroupAttr()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_TUNNEL_E, byref(attr), byref(tunnel_cb)) != 0:
            self._logger.error("Failed to init tunnel drop reason group")

        # acl
        acl_cb = WjhDropCallbacks()
        acl_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ACL_E
        acl_cb.raw_cb.acl = WjhAclDropRawCb(self.wjh_acl_drop_raw_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ACL_E, byref(attr), byref(acl_cb)) != 0:
            self._logger.error("Failed to init acl drop reason group")

        # buffer
        if self.buffer_cb_supported():
            buffer_cb = WjhDropCallbacks()
            buffer_cb.drop_reason_group = WJH_DROP_REASON_GROUP_BUFFER_E
            buffer_cb.raw_cb.buffer = WjhBufferDropRawCb(self.wjh_buffer_drop_raw_cb)
            attr = WjhDropReasonGroupAttr()
            attr.attr.buffer_drop.basic_attr.recirculation_port = 0
            attr.attr.buffer_drop.advanced_attr.trap_probability = 8
            if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_BUFFER_E, byref(attr), byref(buffer_cb)) != 0:
                self._logger.error("Failed to init buffer drop reason group")

            if self.latency_monitor_supported():
                self.init_latency_and_congestion_thresholds()

        # L1
        l1_cb = WjhDropCallbacks()
        l1_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L1_E
        l1_cb.aggregate_cb.L1 = WjhL1DropAggregateCb(self.wjh_l1_drop_aggregate_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L1_E, byref(attr), byref(l1_cb)) != 0:
            self._logger.error("Failed to init L1 drop reason group")

        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E, aggregation_channel_id)

        global client_instance
        client_instance = self

        try:
            with open(SDK_PID_LAST, 'w') as f:
                f.write('{}'.format(self.sdk_pid))
        except Exception as e:
            pass

        check_portmap = 0
        while True:
            time.sleep(15)
            self.flush_agg_cache()
            check_portmap += 1
            if check_portmap == 4:
                check_portmap = 0
                self.check_port_map()


    def check_port_map(self):
        cmd = '/usr/bin/diff {} {}'.format(CL_PORTTAB, CL_PORTTAB_LAST)
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            cmd = '/usr/bin/cp {} {}'.format(CL_PORTTAB, CL_PORTTAB_LAST)
            try:
                output = subprocess_wrapper(cmd.split())
            except Exception:
                pass

            self._logger.info('port map changed')
            self.get_port_map()

    def wjh_deactive_cb(self):
        self._logger.info("This WJH client is deactivated by another WJH client")
        sys.exit(0)


    def get_switchd_mapping_tables(self):
        cmd = 'cat {}/root'.format(SWITCHD_SW_TABLE_PATH)
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            self._logger.error('Cannot get switchd table root')
            return AGENTSTATUS_FAILURE

        for l in output.splitlines():
            if 'Backend Port Info (Linux) Table' in l:
                self.port_map_id = int(l.split()[1])
            if 'Backend Bond Lag Info Table' in l:
                self.lag_map_id = int(l.split()[1])

        if self.port_map_id == 0:
            self.logger.error('Cannot get port map table')
            return AGENTSTATUS_FAILURE

        if self.lag_map_id == 0:
            self.logger.error('Cannot get lag map table')
            return AGENTSTATUS_FAILURE


    def get_port_map(self):
        cmd = 'cat {}/{}'.format(SWITCHD_SW_TABLE_PATH, self.port_map_id)
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            self._logger.error('Cannot get portmap')
            return AGENTSTATUS_FAILURE

        port = ''
        lid = ''
        self.portmap = {}
        self.lidmap = {}
        for l in output.splitlines():
            if 'port-info' in l:
                port = l.split()[1]
            if 'lid' in l:
                lid = l.split()[1]
                self.portmap[lid] = port
                self.lidmap[port] = lid

        return AGENTSTATUS_SUCCESS


    # TBD: the agent process should really have a link cache for efficient
    #      access with up to date information.  Reading from switchd is not
    #      workable as the information can chnage at any time
    '''
    def get_lag_map(self):
        cmd = 'cat {}/{}'.format(SWITCHD_SW_TABLE_PATH, self.lag_map_id)
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            self._logger.error('Cannot get lagmap')
            return AGENTSTATUS_FAILURE

        lag = ''
        lid = ''
        self.lagmap = {}
        self.lidmap = {}
        for l in output.splitlines():
            if 'bond-info-entry' in l:
                lag = l.split()[1]
            if 'lag-id' in l:
                lid = l.split()[1]
                self.lagmap[lid] = lag
                self.lidmap[lag] = lid

        return AGENTSTATUS_SUCCESS
    '''


    def start_dummy_client(self):
        class _raw_info(object):
            def __init__(self, reason='', drop_type='', ingress_port='',
                         egress_port='', traffic_class=0, acl_rule_id=0,
                         acl_bind_point=0, acl_name='', acl_rule='', smac='',
                         dmac='', sip='', dip='', proto=0, sport=0, dport=0,
                         fts=0.0, lts=0.0, agg_count=0, severity='', tc=0,
                         latency=0, occupancy=0):
                self.drop_agg = WjhDropAggregate()
                self.drop_agg.reason = reason
                self.drop_agg.drop_type = drop_type
                self.drop_agg.ingress_port = ingress_port
                self.drop_agg.egress_port = egress_port
                self.drop_agg.traffic_class = traffic_class
                self.drop_agg.acl_rule_id = acl_rule_id
                self.drop_agg.acl_bind_point = acl_bind_point
                self.drop_agg.acl_name = acl_name
                self.drop_agg.acl_rule = acl_rule
                self.drop_agg.smac = smac
                self.drop_agg.dmac = dmac
                self.drop_agg.sip = sip
                self.drop_agg.dip = dip
                self.drop_agg.proto = proto
                self.drop_agg.sport = sport
                self.drop_agg.dport = dport
                self.drop_agg.agg_count = agg_count
                self.drop_agg.first_timestamp = fts
                self.drop_agg.last_timestamp = lts
                self.drop_agg.severity = severity
                if drop_type == 'Buffer':
                    self.drop_agg.traffic_class = tc
                    self.drop_agg.original_latency = latency
                    self.drop_agg.original_occupancy =  occupancy

        self._logger.info('Started DUMMY wjh agent')
        try:
            data_file = WJH_DUMMY_DATA_PATH
            with open(data_file, 'r') as f:
                drop_data = yaml.load(f)
                if drop_data:
                    l1_data = drop_data.get('L1-drops')
                    other_data = drop_data.get('Other-drops')

                    t1 = time.time()
                    if other_data:
                        for ddict in other_data:
                            self._logger.debug('--- Others')
                            d = ddict['drop']
                            self._logger.debug('{}'.format(d))
                            fts = t1 - (random.random() * 20)
                            lts = t1 + (random.random() * 20)

                            raw_info = _raw_info(reason=d.get('reason', ''),
                                                 drop_type=d.get('drop_type', ''),
                                                 ingress_port=d.get('ingress_port', ''),
                                                 egress_port=d.get('egress_port', ''),
                                                 traffic_class=0 if not d.get('traffic_class') else int(d.get('traffic_class')),
                                                 acl_rule_id=0 if not d.get('acl_rule_id') else int(d.get('acl_rule_id')),
                                                 acl_bind_point=0 if not d.get('acl_bind_point') else int(d.get('acl_bind_point')),
                                                 acl_name=d.get('acl_name', ''),
                                                 acl_rule=d.get('acl_rule', ''),
                                                 smac=d.get('smac', ''),
                                                 dmac=d.get('dmac', ''),
                                                 proto=0 if not d.get('proto') else int(d.get('proto')),
                                                 sport=0 if not d.get('sport') else int(d.get('sport')),
                                                 dport=0 if not d.get('dport') else int(d.get('dport')),
                                                 sip=d.get('sip', ''),
                                                 dip=d.get('dip', ''),
                                                 fts=fts,
                                                 lts=lts,
                                                 agg_count=0 if not d.get('count') else int(d.get('count')),
                                                 severity=d.get('severity',''),
                                                 tc=d.get('tc', 0),
                                                 latency=d.get('original_latency', 0),
                                                 occupancy=d.get('original_occupancy', 0))
                            drop_agg = raw_info.drop_agg
                            key = (drop_agg.reason, drop_agg.ingress_port, drop_agg.sip, drop_agg.dip,
                                   drop_agg.proto, drop_agg.sport, drop_agg.dport, drop_agg.smac,
                                   drop_agg.dmac)
                            self.agg_cache[key] = drop_agg
                            self._logger.debug('--- Others saved')
                    if l1_data:
                        for ddict in l1_data:
                            self._logger.debug('--- L1')
                            d = ddict['drop']
                            self._logger.debug('{}'.format(d))
                            fts = t1 - (random.random() * 20)
                            lts = t1 + (random.random() * 20)

                            L1drop = WjhL1DropAggregate()
                            L1drop.ingress_port = d.get('ingress_port', '')
                            L1drop.is_port_up = d.get('is_port_up') == 'True' or 'False'
                            L1drop.port_down_reason = d.get('port_down_reason', '')
                            L1drop.description = ''
                            if d.get('state_change_count'):
                                L1drop.state_change_count = int(d.get('state_change_count'))
                            if d.get('symbol_error_count'):
                                L1drop.symbol_error_count = int(d.get('symbol_error_count'))
                            if d.get('crc_error_count'):
                                L1drop.crc_error_count = int(d.get('crc_error_count'))
                            L1drop.first_timestamp = fts
                            L1drop.last_timestamp = lts
                            L1drop.save()
                            self._logger.debug('--- L1 saved')

        except Exception as e:
            self._logger.error('Error parsing wjh data file: {}'.format(e))

        while True:
            time.sleep(15)
            self.flush_agg_cache()


    def start_wjh_client(self):
        if not self.is_mlx:
            return self.start_dummy_client()

        try:
            ctypes.CDLL('libsxapi.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxadviser.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxaclhelper.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libscew.so', mode=ctypes.RTLD_GLOBAL)
            self.wjh_lib = ctypes.CDLL('libwjh.so', mode=ctypes.RTLD_GLOBAL)
        except OSError:
            try:
                ctypes.CDLL('/usr/lib/x86_64-linux-gnu/libelf.so.1', mode=ctypes.RTLD_GLOBAL)
                self.wjh_lib = ctypes.CDLL('libwjh.so', mode=ctypes.RTLD_GLOBAL)
            except Exception as e:
                self._logger.error("libwjh load error: {}".format(e))
                return AGENTSTATUS_FAILURE

        #wjh_helper = ctypes.CDLL('libwjhhelper.so', mode=ctypes.RTLD_GLOBAL)

        try:
            cmd = '/usr/bin/pidof sx_sdk'
            sdk_pid = subprocess_wrapper(cmd.split())
            self.sdk_pid = int(sdk_pid)

            try:
                with open(SDK_PID_LAST, 'r') as f:
                    old_sdk_pid = int(f.read())

            except Exception:
                old_sdk_pid = 0

            self._logger.info('old sdk pid {}, new sdk pid {}'.format(old_sdk_pid, self.sdk_pid))

            if old_sdk_pid != self.sdk_pid:
                self.need_tele_init = True

        except Exception:
            self._logger.debug("sx_sdk not alive")
            return AGENTSTATUS_FAILURE

        global init_params
        init_param = WjhInitParam()
        init_param.force = 0
        init_param.deactive_cb = WjhDeactiveCb(self.wjh_callback)
        init_param.bandwidth_percent = 0
        init_param.log_cb = 0
        init_param.conf_xml_path = c_char_p(WJH_LIB_XML_PATH.encode('utf-8'))
        init_param.debug_fs_path = 0
        init_param.ingress_info_type = WJH_INGRESS_INFO_TYPE_LOGPORT


        if self.wjh_lib.wjh_init(byref(init_param)) != 0:
            self._logger.error("wjh_init failed")
            return AGENTSTATUS_FAILURE

        channel_id = self.create_user_channel(WJH_USER_CHANNEL_CYCLIC_E)
        aggregation_channel_id = self.create_user_channel(WJH_USER_CHANNEL_AGGREGATE_E)

        self.wjh_callback(channel_id, aggregation_channel_id)

        self._logger.info('WJH DONE, EXITING')
        if self.wjh_lib.wjh_user_channel_destroy(channel_id) != 0:
            self._logger.error("Failed to destroy user channel")

        if self.wjh_lib.wjh_user_channel_destroy(aggregation_channel_id) != 0:
            self._logger.error("Failed to destroy agg user channel")

        if self.wjh_lib.wjh_deinit() != 0:
            self._logger.error("Failed to deinit wjh service lib")

        self._logger.info("wjh lib is deinitialized")
        return AGENTSTATUS_SUCCESS


    #
    # Raw callback handlers
    #
    def wjh_raw_drop_fill(self, info):
        drop_reason = WjhDropRawInfo()
        drop_reason.reason_id = info.drop_reason.id
        drop_reason.reason = info.drop_reason.reason.decode('utf-8')
        drop_reason.description = info.drop_reason.description
        drop_reason.severity = severity_to_str.get(info.drop_reason.severity) or ''
        lid = '0x%x' % info.ingress_port
        drop_reason.ingress_port = self.portmap.get(lid) or lid
        drop_reason.timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)

        packet_size = info.packet_size
        packet = list((packet_size * ctypes.c_ubyte).from_address(info.packet))

        drop_reason.packet_size = packet_size
        if (packet_size > 0):
            drop_reason.packet = ''.join('%02x' % i for i in packet)

        return drop_reason

    def __parse_tuple(self, eth, tup):
        tup['smac'] = eth.src
        tup['dmac'] = eth.dst

        l3_found = False
        iph = None

        try:
            iph = eth[IP]
            l3_found = True
            tup['proto'] = iph.proto
        except Exception as e:
            self._logger.debug('{}'.format(e))

        if not l3_found:
            try:
                iph = eth[IPv6]
                l3_found = True
                tup['proto'] = iph.nh
            except Exception as e:
                self._logger.debug('{}'.format(e))

        if l3_found:
            tup['src'] = iph.src
            tup['dst'] = iph.dst

            try:
                udph = eth[UDP]
                tup['sport'] = udph.sport
                tup['dport'] = udph.dport

            except Exception as e:
                self._logger.debug('{}'.format(e))

            try:
                tcph = eth[TCP]
                tup['sport'] = tcph.sport
                tup['dport'] = tcph.dport
            except Exception as e:
                self._logger.debug('{}'.format(e))

        return tup

    def parse_7_tuples(self, packet, packet_size):
        tup = {'smac': '00:00:00:00:00:00',
               'dmac': '00:00:00:00:00:00',
               'src': '0.0.0.0',
               'dst': '0.0.0.0',
               'sport': 0,
               'dport': 0,
               'proto': 0}

        inner_tup = None

        if packet_size == 0:
            return tup

        pkt = list((packet_size * ctypes.c_ubyte).from_address(packet))
        try:
            p = raw(binascii.a2b_hex(''.join('%02x' % i for i in pkt)))
            eth = Ether(p)
        except Exception as e:
            return tup

        tup = self.__parse_tuple(eth, tup)
        if tup and tup['proto'] == 17 and tup['dport'] == 4789:
            vxlan = eth[VXLAN]
            ieth = vxlan[Ether]
            inner_tup = {'smac': '00:00:00:00:00:00',
                         'dmac': '00:00:00:00:00:00',
                         'src': '0.0.0.0',
                         'dst': '0.0.0.0',
                         'sport': 0,
                         'dport': 0,
                         'proto': 0}

            inner_tup = self.__parse_tuple(ieth, inner_tup)

        self._logger.debug('smac: {}, dmac: {} -- src: {}, dst: {}, proto {}, sport {}, dport {}'.format(
                           tup['smac'], tup['dmac'], tup['src'], tup['dst'], tup['proto'], tup['sport'], tup['dport']))
        return tup, inner_tup


    def __fill_event_tuples(self, event, tuples, info):
        event.sip = tuples['src']
        event.dip = tuples['dst']
        event.proto = tuples['proto']
        event.sport = tuples.get('sport', 0)
        event.dport = tuples.get('dport', 0)
        event.smac = tuples['smac']
        event.dmac = tuples['dmac']
        key = (event.reason, event.ingress_port, event.sip, event.dip,
               event.proto, event.sport, event.dport, event.smac,
               event.dmac)
        self._logger.debug('ADD key: {}'.format(key))

        agg = self.agg_cache.get(key)
        if agg:
            agg.agg_count += 1
            agg.last_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
        else:
            event.agg_count = 1
            event.first_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
            event.last_timestamp = event.first_timestamp
            self.agg_cache[key] = event

        return event

    def wjh_raw_drop_fill_aggregate(self, drop_type, info, egress_port='',
                                    traffic_class=0, acl_rule_id=0,
                                    original_occupancy=0, original_latency=0):
        drop_reason = WjhDropAggregate()
        if drop_type == 'ACL':
            if info.binding_point == WJH_ACL_BINDING_POINT_INGRESS and info.acl_name.decode('utf-8') == 'monitor' and info.rule.decode('utf-8') == '-':
                r = 'Monitor'
            else:
                r = acl_bp_to_str.get(info.binding_point) or ''
            drop_reason.reason = '{}'.format(r)
            drop_reason.acl_bind_point = info.binding_point
            drop_reason.acl_name = info.acl_name.decode('utf-8') if info.acl_name else ''
            drop_reason.acl_rule = info.rule.decode('utf-8') if info.rule else ''
        else:
            drop_reason.reason = info.drop_reason.reason.decode('utf-8')
        drop_reason.drop_type = drop_type
        drop_reason.severity = severity_to_str.get(info.drop_reason.severity) or ''
        drop_reason.egress_port = egress_port
        drop_reason.acl_rule_id = acl_rule_id
        if traffic_class != WJH_MIRROR_TCLASS_INVALID:
            drop_reason.traffic_class = traffic_class
        if original_occupancy != WJH_MIRROR_CONG_INVALID:
            drop_reason.original_occupancy = max(original_occupancy, drop_reason.original_occupancy)
        if original_latency != WJH_MIRROR_LATENCY_INVALID:
            drop_reason.original_latency = max(original_latency, drop_reason.original_latency) * WJH_MIRROR_LATENCY_NSEC_UNIT
        lid = '0x%x' % info.ingress_port
        drop_reason.ingress_port = self.portmap.get(lid) or lid


        p_tuples, p_inner_tuples = self.parse_7_tuples(info.packet, info.packet_size)
        self.__fill_event_tuples(drop_reason, p_tuples, info)
        if p_inner_tuples:
            inner_event = WjhDropAggregate()
            inner_event.drop_type = drop_reason.drop_type
            inner_event.reason = drop_reason.reason
            inner_event.acl_bind_point = drop_reason.acl_bind_point
            if drop_reason.reason == 'Monitor':
                inner_event.acl_name = 'tunneled'
            inner_event.acl_rule_id = drop_reason.acl_rule_id
            inner_event.severity = drop_reason.severity
            inner_event.egress_port = drop_reason.egress_port
            inner_event.traffic_class = drop_reason.traffic_class
            inner_event.original_occupancy = drop_reason.original_occupancy
            inner_event.original_latency = drop_reason.original_latency
            inner_event.ingress_port = drop_reason.ingress_port
            self.__fill_event_tuples(inner_event, p_inner_tuples, info)

        return drop_reason

    def flush_agg_cache(self):
        for key, agg in list(self.agg_cache.items()):
            self._logger.info("Aggregate type: {}, count: {} first {} last {}".format(agg.drop_type, agg.agg_count, agg.first_timestamp, agg.last_timestamp))
            agg.save()

        self.agg_cache.clear()
        mcls = WjhDropAggregate.memmodel()
        mcls.flush()


    def wjh_l2_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if raw_info_list_p[i].drop_reason.id in default_filters:
                continue
            if self.is_wjh_drop_filter_configured(raw_info_list_p[i], WjhConfKw.L2):
                continue
            self._logger.debug("l2 drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str.get(raw_info_list_p[i].drop_reason.severity) or '',
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('L2', raw_info_list_p[i])

        return 0


    def wjh_router_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if raw_info_list_p[i].drop_reason.id in default_filters:
                continue
            if self.is_wjh_drop_filter_configured(raw_info_list_p[i], WjhConfKw.ROUTER):
                continue
            self._logger.debug("router drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str.get(raw_info_list_p[i].drop_reason.severity) or '',
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Router', raw_info_list_p[i])

        return 0


    def wjh_tunnel_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if self.is_wjh_drop_filter_configured(raw_info_list_p[i], WjhConfKw.TUNNEL):
                continue
            self._logger.debug("tunnel drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str.get(raw_info_list_p[i].drop_reason.severity) or '',
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Tunnel', raw_info_list_p[i])

        return 0


    def wjh_buffer_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if self.is_wjh_drop_filter_configured(raw_info_list_p[i], WjhConfKw.BUFFER):
                continue
            self._logger.debug("buffer drop reason group: ingress port [0x%x], reason[%s], egress port [%d: 0x%x], tc [%d], occupancy[%d], latency[%d], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].egress_port_valid,
                      raw_info_list_p[i].egress_port,
                      raw_info_list_p[i].tc,
                      raw_info_list_p[i].original_occupancy,
                      raw_info_list_p[i].original_latency,
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            if raw_info_list_p[i].egress_port_valid:
                egress_port = raw_info_list_p[i].egress_port
            else:
                egress_port = ''
            self.wjh_raw_drop_fill_aggregate('Buffer', raw_info_list_p[i],
                                   egress_port=egress_port,
                                   traffic_class=raw_info_list_p[i].tc,
                                   original_occupancy=raw_info_list_p[i].original_occupancy * WJH_BUFFER_OCCUPANCY_UNIT,
                                   original_latency=raw_info_list_p[i].original_latency * WJH_MIRROR_LATENCY_NSEC_UNIT)

        return 0


    def wjh_acl_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if self.is_wjh_drop_filter_configured(raw_info_list_p[i], WjhConfKw.ACL):
                continue
            self._logger.debug("acl drop reason group: ingress port [0x%x], size[%d], rule_id[0x%x], binding_point[0x%x], acl_name[%s], rule{%s}, timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                     raw_info_list_p[i].packet_size,
                     raw_info_list_p[i].rule_id,
                     raw_info_list_p[i].binding_point,
                     raw_info_list_p[i].acl_name,
                     raw_info_list_p[i].rule,
                     raw_info_list_p[i].timestamp.tv_sec,
                     raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('ACL', raw_info_list_p[i], acl_rule_id=raw_info_list_p[i].rule_id)
        return 0


    #
    # Aggregate callback handlers
    #
    def wjh_agg_drop_fill(self, key, data):
        drop = WjhDropAggregate()
        lid = '0x%x' % key.ingress_port
        drop.ingress_port = self.portmap.get(lid) or lid
        if key.is_lag_member:
            drop.ingress_lag = key.ingress_lag
        else:
            drop.ingress_lag = ''
        drop.sip = key.five_tuples.sip
        drop.dip = key.five_tuples.dip
        drop.proto = key.five_tuples.proto
        drop.sport = key.five_tuples.sport
        drop.dport = key.five_tuples.dport
        drop.smac = '%02x:%02x:%02x:%02x:%02x:%02x' % (key.smac[0], key.smac[1], key.smac[2],
                                                       key.smac[3], key.smac[4], key.smac[5])
        drop.dmac = '%02x:%02x:%02x:%02x:%02x:%02x' % (key.dmac[0], key.dmac[1], key.dmac[2],
                                                       key.dmac[3], key.dmac[4], key.dmac[5])
        # only in agg cb, not included in netq model
        #drop.vlan = key.vlan
        #drop.ether_type = key.ether_type
        drop.agg_count = data.count
        drop.first_timestamp = float(data.timestamp.first_timestamp.tv_sec) + float(
                                     data.timestamp.first_timestamp.tv_nsec / 1000000000.0)
        drop.last_timestamp = float(data.timestamp.last_timestamp.tv_sec) + float(
                                           data.timestamp.last_timestamp.tv_nsec / 1000000000.0)
        return drop


    def wjh_acl_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p=None):
        # TBD: bind point is not available
        list_size = cnt_p.contents.value
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'ACL'
            drop_reason.acl_rule_id = key.rule_id
            drop_reason.acl_name = key.acl_name
            drop_reason.acl_rule = key.rule
            # drop_reason.acl_bind_point = acl_bp_to_str.get(info.binding_point) or '' # not available with agg cb
            drop_reason.save()

    def wjh_buf_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p=None):
        # buffer drop aggregate is not useful as it does not have traffic_class
        # info
        pass

    def wjh_l1_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        for i in range(list_size):
            lid = '0x%x' % key_list_p[i].ingress_port
            ingress_port = self.portmap.get(lid) or lid
            '''
            self._logger.info("port %s: is_port_up: %u, port_down_reason: %s, description: %s, state_change_count: %u" %
                  (ingress_port, data_list_p[i].is_port_up, data_list_p[i].port_down_reason,
                   data_list_p[i].description, data_list_p[i].state_change_count))
            self._logger.info("           symbol_error_count: %u, crc_error_count: %u" %
                  (data_list_p[i].symbol_error_count, data_list_p[i].crc_error_count))
            self._logger.info("           timestamp[%ld.%ld, %ld.%ld]" % (data_list_p[i].timestamp.first_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.first_timestamp.tv_nsec, data_list_p[i].timestamp.last_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.last_timestamp.tv_nsec))
            '''
            # L1 drop callback data is quite different from the others
            filter_drop_reasons, filter_drop_severity = self.get_wjh_filter_config(WjhConfKw.L1)
            if data_list_p[i].port_down_reason.decode('utf-8') in filter_drop_reasons:
                self._logger.debug('L1 drop with reason {} filtered out'
                                   .format(data_list_p[i].port_down_reason.decode('utf-8')))
                continue

            l1drop = WjhL1DropAggregate()
            l1drop.ingress_port = ingress_port
            l1drop.is_port_up = data_list_p[i].is_port_up == 1
            l1drop.port_down_reason = data_list_p[i].port_down_reason.decode('utf-8')
            l1drop.description = data_list_p[i].description
            l1drop.state_change_count = data_list_p[i].state_change_count
            l1drop.symbol_error_count = data_list_p[i].symbol_error_count
            l1drop.crc_error_count = data_list_p[i].crc_error_count
            l1drop.first_timestamp = float(data_list_p[i].timestamp.first_timestamp.tv_sec) + float(
                                           data_list_p[i].timestamp.first_timestamp.tv_nsec / 1000000000.0)
            l1drop.last_timestamp = float(data_list_p[i].timestamp.last_timestamp.tv_sec) + float(
                                          data_list_p[i].timestamp.last_timestamp.tv_nsec / 1000000000.0)
            olddrops = WjhL1DropAggregate().memmodel()
            olddrop = olddrops.query.get(hostname=HOSTNAME, ingress_port=ingress_port)
            if (olddrop and olddrop.is_port_up == l1drop.is_port_up and
                olddrop.port_down_reason == l1drop.port_down_reason and
                olddrop.state_change_count == l1drop.state_change_count and
                olddrop.symbol_error_count == l1drop.symbol_error_count and
                olddrop.crc_error_count == l1drop.crc_error_count):
                continue

            l1drop.save()

        return 0


    def wjh_rtr_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'Router'
            drop_reason.reason = key.reason.decode('utf-8')
            drop_reason.save()


    def wjh_l2_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'L2'
            drop_reason.reason = key.reason.decode('utf-8')
            drop_reason.save()


    def wjh_tun_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'Tunnel'
            drop_reason.reason = key.reason.decode('utf-8')
            drop_reason.save()

    def get_wjh_filter_config(self, filter_type):
        wjh_fiter_list = self._conf.get('netq-agent').get('wjh_filter', [])
        for wjh_filter in wjh_fiter_list:
            if filter_type in wjh_filter.keys():
                filter_type_config = wjh_filter.get(filter_type, {})
                drop_reason_dict = WjhConfKw.DROP_REASONS_MAPPING.get(filter_type)
                configured_drop_reasons = filter_type_config.get('drop-reasons', [])
                filter_drop_reasons = [drop_reason_dict.get(reason) for reason in configured_drop_reasons]
                filter_drop_severity = filter_type_config.get('severity', [])
                return filter_drop_reasons, filter_drop_severity
        return [], []

    def is_wjh_drop_filter_configured(self, raw_info, drop_type):
        filter_drop_reasons, filter_drop_severity = self.get_wjh_filter_config(drop_type)
        drop_reason= raw_info.drop_reason.reason.decode('utf-8')
        if drop_reason in filter_drop_reasons:
            self._logger.debug('{} drop with reason {} filtered'.format(drop_type, drop_reason))
            return True
        if severity_to_str.get(raw_info.drop_reason.severity) in filter_drop_severity:
            self._logger.debug('{} drop with severity {} filtered'.format(drop_type,
                                                                          severity_to_str.get(raw_info.drop_reason.severity)))
            return True

        return False

    def _run():
        pass

    def clean_up(self):
        self._logger.info('WJH clean up')
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E)


def sig_handler(signum, frame):
    if client_instance:
        client_instance.clean_up()
    sys.exit(0)
