# Copyright 2017, 2019, 2020 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import re

from netq_lib.orm.redisdb.models import (
    Temp,
    Fan,
    PSU
)
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.sensors import SensorRender, SensorPowerRender, FAIL_TEMP, FAIL_FAN


class SonicSensorRender(SensorRender):
    def __init__(self):
        super(SonicSensorRender, self).__init__()

    def process_fan_sensors(self, fullstate, logger):
        fan_arr = []
        result = self.run_command("show platform fan")
        if result.success and result.op:
            #         FAN     Speed    Direction     Presence    Status          Timestamp
            # -----------  --------  -----------  -----------  --------  -----------------
            header = "        FAN     Speed    Direction     Presence    Status          Timestamp".split()
            idx = 0
            matches_positions = []
            table_margin = 0
            for row in result.op.decode().split("\n"):
                if idx == 0:
                    if header != row.split():
                        logger.error("Command 'show platform fan' header validation failed")
                        return
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif len(row) != table_margin:
                    logger.warn("Invalid row {}".format(row))
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].lstrip() for i in range(len(matches_positions)-1)]
                    name = data[0]
                    status = data[4]
                    min_speed = FAIL_FAN
                    max_speed = FAIL_FAN
                    current_speed = FAIL_FAN
                    r_min_speed = run_command("cat /var/run/hw-management/thermal/{}_min".format(name))
                    if r_min_speed.success and r_min_speed.op:
                        min_speed = int(r_min_speed.op.decode().strip())
                    r_max_speed = run_command("cat /var/run/hw-management/thermal/{}_max".format(name))
                    if r_max_speed.success and r_max_speed.op:
                        max_speed = int(r_max_speed.op.decode().strip())
                    r_current_speed = run_command("cat /var/run/hw-management/thermal/{}_speed_get".format(name))
                    if r_current_speed.success and r_current_speed.op:
                        current_speed = int(r_current_speed.op.decode().strip())
                    sensor = Fan()
                    sensor.s_name = name
                    sensor.s_state = status
                    sensor.s_input = current_speed
                    sensor.s_max = max_speed
                    sensor.s_min = min_speed

                    sensor.save()
                    if fullstate:
                        fan_arr.append(sensor)

                idx += 1

            if fullstate:
                save_fullstate(fan_arr, Fan.__name__)


    def process_temp_sensors(self, fullstate, logger):
        temp_arr = []
        result = self.run_command("show platform temperature")
        if result.success and result.op:
            #                Sensor    Temperature    High TH    Low TH    Crit High TH    Crit Low TH    Warning          Timestamp
            # ---------------------  -------------  ---------  --------  --------------  -------------  ---------  -----------------
            header = "               Sensor    Temperature    High TH    Low TH    Crit High TH    Crit Low TH    Warning          Timestamp".split()
            idx = 0
            matches_positions = []
            table_margin = 0
            for row in result.op.decode().split("\n"):
                if idx == 0:
                    if header != row.split():
                        logger.error("Command 'show platform temperature' header validation failed")
                        return
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif len(row) != table_margin:
                    logger.warn("Invalid row - {}".format(row))
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].lstrip() for i in range(len(matches_positions)-1)]
                    sensor = Temp()
                    sensor.s_name = data[0]
                    sensor.s_input = float(data[1]) if data[1] != "N/A" else float(FAIL_TEMP)
                    sensor.s_max = int(float(data[2])) if data[2] != "N/A" else FAIL_TEMP
                    sensor.s_min = int(float(data[3])) if data[3] != "N/A" else FAIL_TEMP
                    sensor.s_crit = int(float(data[4])) if data[4] != "N/A" else FAIL_TEMP
                    sensor.s_lcrit = int(float(data[5])) if data[5] != "N/A" else FAIL_TEMP

                    sensor.save()

                    if fullstate:
                        temp_arr.append(sensor)
                idx += 1
            if fullstate:
                save_fullstate(temp_arr, Temp.__name__)


    def process_psu_sensors(self, fullstate, logger):
        psu_arr = []
        result = self.run_command("show platform psustatus")
        if result.success and result.op:
            # PSU    Status
            # -----  --------
            header = "PSU    Status".split()
            idx = 0
            matches_positions = []
            table_margin = 0
            for row in result.op.decode().split("\n"):
                if idx == 0:
                    if header != row.split():
                        logger.error("Command 'show platform psustatus' header validation failed")
                        return
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif not len(row):
                    logger.warn("Invalid row")
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].strip() for i in range(len(matches_positions)-1)]
                    sensor = PSU()
                    sensor.s_name = data[0]
                    sensor.s_state = data[1]

                    sensor.save()
                    if fullstate:
                        psu_arr.append(sensor)
                idx += 1
            if fullstate:
                save_fullstate(psu_arr, PSU.__name__)

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get('fullstate')
        logger = kwargs.get('logger')
        if svc_failed:
            return ServiceStatus.FAIL

        self.process_fan_sensors(fullstate, logger)
        self.process_temp_sensors(fullstate, logger)
        self.process_psu_sensors(fullstate, logger)
        return ServiceStatus.OK


def process_sensors(cmd_output, svc_failed=False, **kwargs):
    render = SonicSensorRender()
    return render.parse(cmd_output, svc_failed, **kwargs)


class SonicSensorPowerRender(SensorPowerRender):
    def __init__(self):
        super(SonicSensorPowerRender, self).__init__()


def process_power(cmd_output, svc_failed=False, **kwargs):
    render = SonicSensorPowerRender()
    return render.parse(cmd_output, svc_failed, **kwargs)

